/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.kernel.configuration;

import java.util.Set;

import org.ow2.petals.kernel.ws.api.to.ContainerInformation;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.generated.Topology;

/**
 * The interface of the Configuration service. The service handles the
 * configuration of the local PEtALS container.<br>
 * At its starting, it retrieves the configuration from the
 * {@link TopologyBuilder} parser and the server.properties file , it performs a
 * first level validation of the consistency, then it fill the local
 * {@link ContainerInformation}.<br>
 * It provides a method to validate an new container against the topology.
 * 
 * @author Roland NAUDIN, chamerling - EBMWebsourcing
 * @author Charles Casadei - eBM WebSourcing
 * @since Petals 2.0
 */
public interface ConfigurationService {

    /**
     * Path of the default topology file.
     */
    static final String TOPOLOGY_FILE = "topology.xml";

    static final String TOPOLOGY_RESOURCE = "/" + TOPOLOGY_FILE;

    /**
     * Name of the server properties file.
     */
    static final String SERVER_PROPS_FILE = "/server.properties";

    static final String JAAS_PROPS_FILE = "/jaas.properties";

    /**
     * Properties keys
     */
    static final String PROPERTY_CONTAINER_NAME = "petals.container.name";

    static final String PROPERTY_REPOSITORY_PATH = "petals.repository.path";

    static final String PROPERTY_TOPOLOGY_URL = "petals.topology.url";

    static final String PROPERTY_EXCHANGE_VALIDATION = "petals.exchange.validation";

    static final String PROPERTY_ROUTER_STRATEGY = "petals.router.strategy";
    
    static final String PROPERTY_ROUTER_SEND_ATTEMPT = "petals.router.send.attempt";
    
    static final String PROPERTY_ROUTER_SEND_DELAY = "petals.router.send.delay";

    static final String PROPERTY_TASK_TIMEOUT = "petals.task.timeout";

    static final String PROPERTY_CLASSLOADER_ISOLATED = "petals.classloaders.isolated";

    static final String PROPERTY_AUTOLOADER = "petals.autoloader";

    static final String PROPERTY_SSL_KEY_PASSWORD = "petals.ssl.key.password";

    static final String PROPERTY_SSL_KEYSTORE_FILE = "petals.ssl.keystore.file";

    static final String PROPERTY_SSL_KEYSTORE_PASSWORD = "petals.ssl.keystore.password";

    static final String PROPERTY_SSL_TRUSTSTORE_FILE = "petals.ssl.truststore.file";

    static final String PROPERTY_SSL_TRUSTSTORE_PASSWORD = "petals.ssl.truststore.password";
    
    static final String PROPERTY_TRANSPORT_QOS = "petals.transport.qos";

    static final String PROPERTY_TRANSPORT_TCP_RECEIVERS = "petals.transport.tcp.receivers";

    static final String PROPERTY_TRANSPORT_TCP_SENDERS = "petals.transport.tcp.senders";

    static final String PROPERTY_TRANSPORT_TCP_CONNECTION_TIMEOUT = "petals.transport.tcp.connection.timeout";

    static final String PROPERTY_TRANSPORT_TCP_SEND_TIMEOUT = "petals.transport.tcp.send.timeout";

    static final String PROPERTY_TRANSPORT_TCP_SEND_EVICTOR_DELAY = "petals.transport.tcp.send.evictor.deplay";

    static final String PROPERTY_TRANSPORT_TCP_SEND_EVICTABLE_DELAY = "petals.transport.tcp.send.evictable.deplay";

    static final String PROPERTY_USER_CONFIG_PREFIX = "user";

    static final String PROPERTY_REGISTRY_TRANSPORT_TIMEOUT = "registry.transport.timeout";

    static final String PROPERTY_REGISTRY_SYNCHRO_PERIOD = "registry.synchro.period";

    static final String PROPERTY_REGISTRY_INCOMING_MANAGER = "registry.incoming.manager";

    static final String PROPERTY_REGISTRY_OUTGOING_MANAGER = "registry.outgoing.manager";

    static final String PROPERTY_REGISTRY_REPOSITORY = "registry.repository";

    static final String PROPERTY_REGISTRY_MESSAGE_RECEIVER = "registry.message.receiver";

    static final String PROPERTY_REGISTRY_MESSAGE_SENDER = "registry.message.sender";

    static final String PROPERTY_REGISTRY_CLIENT_MANAGER = "registry.client.manager";

    static final String PROPERTY_REGISTRY_CLIENT_MANAGER_URL = "registry.client.manager.url";

    static final String PROPERTY_REGISTRY_DB_DRIVER = "registry.db.driver";

    static final String PROPERTY_REGISTRY_DB_URL = "registry.db.url";

    static final String PROPERTY_REGISTRY_DB_USER = "registry.db.user";

    static final String PROPERTY_REGISTRY_DB_PASSWORD = "registry.db.password";

    static final String PROPERTY_REGISTRY_DB_DIALECT = "registry.db.dialect";

    static final String PROPERTY_REGISTRY_MODE = "registry.mode";

    static final String STANDALONE = "standalone";

    static final String PEER = "peer";

    static final String MASTER = "master";

    static final String SLAVE = "slave";

    static final String TOPOLOGY_UPDATE_PERIOD = "topology.update.period";
    
    /**
     * Get the local container configuration.
     * 
     */
    ContainerConfiguration getContainerConfiguration();

    /**
     * Get the containers configuration. This only returns the container
     * configuration created at startup from local persisted data.
     * 
     * @return
     */
    public Set<ContainerConfiguration> getContainersConfiguration();

    /**
     * Get the subdomains configuration. This only returns the subdomain
     * configuration created at startup from the local persisted data.
     * 
     * @return
     */
    public Set<SubDomainConfiguration> getSubDomainsConfiguration();

    /**
     * Get the sub domain configuration the local container belongs to.
     * 
     * @return
     */
    SubDomainConfiguration getSubDomainConfiguration();

    /**
     * Get the domain configuration the local container belongs to.
     * 
     * @return
     */
    DomainConfiguration getDomainConfiguration();

    /**
     * Get the current topology. This is the topology loaded from configuration
     * files.
     * 
     * @return
     */
    Topology getTopology() throws ConfigurationException;

    /**
     * Load or reload the topology.
     * 
     * @throws ConfigurationException
     */
    void loadTopology() throws ConfigurationException;

    /**
     * Set the running state of the local container.
     * 
     * @param state
     *            {@code ContainerConfiguration.ContainerState}
     */
    void setContainerState(ContainerConfiguration.ContainerState state);

    /**
     * Verify the given {@code containerConfiguration} against the topology
     * configuration.
     * 
     * @param newContainerConfiguration
     *            The new container configuration
     * @param newSubDomainConfiguration
     *            The new sub-domain configuration
     * @param newDomainConfiguration
     *            The new domain configuration
     * @throws TopologyException
     *             Inconsistency between the topology configuration and the new
     *             container configuration
     */
    void verifyContainerTopology(final ContainerConfiguration newContainerConfiguration,
            final SubDomainConfiguration newSubDomainConfiguration,
            final DomainConfiguration newDomainConfiguration) throws TopologyException;

    /**
     * Get the container configuration.
     * 
     * @param containerName
     *            the name of the container to get configuration from. If null,
     *            returns the local container configuration.
     * @return
     */
    ContainerConfiguration getContainerConfiguration(String containerName);
    
    /**
     * Registers a new container configuration within the local configuration.
     *
     * @param addedContainer the container to be registered
     * @throws ConfigurationException
     */
    void addContainerConfiguration(ContainerConfiguration addedContainer) throws ConfigurationException;

    /**
     * Unregisters the subdomain within the local topology.
     *
     * @param removedContainer
     * @throws ConfigurationException
     */
    void removeContainerConfiguration(ContainerConfiguration removedContainer) throws ConfigurationException;

    /**
     * Registers a new subdomain within the local topology.
     * 
     * @param addedSubdomain
     * @throws ConfigurationException
     */
    void addSubdomainConfiguration(SubDomainConfiguration addedSubdomain) throws ConfigurationException;
    
    /**
     * Unregister a subdomain within the local topology.
     *
     * @param removedSubdomain
     * @throws ConfigurationException
     */
    void removeSubdomainConfiguration(String removedSubdomain) throws ConfigurationException;

    /**
     * Validate the dynamic topology
     *
     * @throws ConfigurationException
     */
    void validateDynamicTopology() throws ConfigurationException;

    /**
     * Backups the local topology.
     *
     * @throws ConfigurationException
     */
    void backupTopology() throws ConfigurationException;
}
