/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.kernel.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.bind.JAXBException;

import org.apache.commons.io.FileUtils;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.routing.RouterService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.NodeType;
import org.ow2.petals.topology.generated.RegistryService;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.SubdomainMode;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;
import org.ow2.petals.topology.generated.WebServiceService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.StringHelper;

/**
 * Class implementing the {@link ConfigurationService} interface.<br>
 * FIXME : Create an implementation for distributed mode and one for standalone
 * mode. This will avoid some if then tests.
 * 
 * @author Roland NAUDIN, chamerling - EBMWebsourcing
 * @author Charles Casadei - eBM WebSourcing
 * @since Petals 2.0
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = ConfigurationService.class))
public class ConfigurationServiceImpl implements ConfigurationService {
    
    private class SubDomainAndContainerNames {

        protected String subDomainName;
        protected String containerName;
    }
    /**
     * Logging utilities
     */
    protected LoggingUtil log;
    @Monolog(name = "logger")
    protected Logger logger;
    // Local informations
    private ContainerConfiguration containerConfiguration;
    private SubDomainConfiguration subDomainConfiguration;
    private DomainConfiguration domainConfiguration;
    private Properties serverProperties;
    private Topology topology;

    // global informations

    /**
     * This is the initial containers configuration created from the topology
     * file. Key is the container name.
     */
    private Map<String, ContainerConfiguration> containersConfig;
    
    /**
     * The subdomains configuration created from the topology file. Key is the
     * subdomain name.
     */
    private Map<String, SubDomainConfiguration> subDomainsConfig;

    /*
     * (non-Javadoc)
     *
     * @seeorg.ow2.petals.kernel.configuration.ConfigurationService#
     * getContainerConfiguration()
     */
    public ContainerConfiguration getContainerConfiguration() {
        return this.containerConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void addContainerConfiguration(ContainerConfiguration addedContainer) throws ConfigurationException {

        if (addedContainer == null) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't add a null container to the local topology");
            throw new ConfigurationException(errorSB.toString());
        }

        final String localContainerName = this.containerConfiguration.getName();

        //Retrieve the container name
        String addedContainerName = null;
        if (addedContainer.getName() == null) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't retrieve the container name");
            throw new ConfigurationException(errorSB.toString());
        }
        addedContainerName = addedContainer.getName();

        if (addedContainerName.equals(localContainerName)) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't add a container with the same name as the current one");
            throw new ConfigurationException(errorSB.toString());
        }

        String addedContainerSubdomainName = addedContainer.getSubdomainName();
        if (addedContainerSubdomainName == null) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("Null subdomain found for the added container : " + addedContainer);
            throw new ConfigurationException(errorSB.toString());
        }

        if (this.containersConfig.containsKey(addedContainer.getName())) {
            this.removeContainerConfiguration(addedContainer);
        }

        //Adds the container to the "JAXB Topology" file
        List<Subdomain> jaxbSubdomains = this.topology.getDomain().getSubDomain();;
        Subdomain jaxbSubdomainOfTheAddedContainer = null;
        for (Subdomain jaxbSubdomain : jaxbSubdomains) {
            if (jaxbSubdomain.getName().equals(addedContainerSubdomainName)) {
                jaxbSubdomainOfTheAddedContainer = jaxbSubdomain;
                break;
            }
        }

        if (jaxbSubdomainOfTheAddedContainer == null) {
            StringBuilder errorSB = new StringBuilder();
            errorSB.append("No subdomain found for the added container : " + addedContainer);
            throw new ConfigurationException(errorSB.toString());
        }
        
        Container addedJaxbContainer = new Container();
        addedJaxbContainer.setName(addedContainer.getName());
        addedJaxbContainer.setDescription(addedContainer.getDescription());
        addedJaxbContainer.setHost(addedContainer.getHost());
        addedJaxbContainer.setUser(addedContainer.getUser());
        addedJaxbContainer.setPassword(addedContainer.getPassword());

        JmxService jaxbJmxService = new JmxService();
        jaxbJmxService.setRmiPort(addedContainer.getJmxRMIConnectorPort());
        addedJaxbContainer.setJmxService(jaxbJmxService);

        WebServiceService jaxbWebService = new WebServiceService();
        jaxbWebService.setPort(addedContainer.getWebservicePort());
        jaxbWebService.setPrefix(addedContainer.getWebservicePrefix());
        addedJaxbContainer.setWebserviceService(jaxbWebService);

        RegistryService jaxbRegistryService = new RegistryService();
        jaxbRegistryService.setPort(addedContainer.getRegistryPort());
        addedJaxbContainer.setRegistryService(jaxbRegistryService);

        TransportService jaxbTransportService = new TransportService();
        jaxbTransportService.setTcpPort(addedContainer.getTCPPort());
        addedJaxbContainer.setTransportService(jaxbTransportService);

        switch (addedContainer.getRegistryMode()) {
            case PEER :
                addedJaxbContainer.setType(NodeType.PEER);
                break;
            case MASTER :
                addedJaxbContainer.setType(NodeType.MASTER);
                break;
            case SLAVE :
                addedJaxbContainer.setType(NodeType.SLAVE);
                break;
            case STANDALONE :
                addedJaxbContainer.setType(NodeType.STANDALONE);
                break;
            default :
                throw new ConfigurationException("Unknown registry mode");
        }

        jaxbSubdomainOfTheAddedContainer.getContainer().add(addedJaxbContainer);
        
        //Store the new container to the topology stored in memory
        this.containersConfig.put(addedContainerName, addedContainer);
    }

    public synchronized void removeContainerConfiguration(ContainerConfiguration removedContainer) throws ConfigurationException {
        if (removedContainer == null) {
            throw new ConfigurationException("Can't remove container : null");
        }

        if (! this.containersConfig.containsKey(removedContainer.getName())) {
            throw new ConfigurationException("No container found for name : " + removedContainer.getName());
        }

        String removedContainerSubdomainName = removedContainer.getSubdomainName();
        if (removedContainerSubdomainName == null) {
            throw new ConfigurationException("No subdomain name found for the removed container");
        }
        
        List<Subdomain> jaxbSubdomains = this.topology.getDomain().getSubDomain();
        Subdomain jaxbSubdomainOfTheRemovedContainer = null;

        for (Subdomain jaxbSubdomain : jaxbSubdomains) {
            if (jaxbSubdomain.getName().equals(removedContainerSubdomainName)) {
                jaxbSubdomainOfTheRemovedContainer = jaxbSubdomain;
                break;
            }
        }

        if (jaxbSubdomainOfTheRemovedContainer == null) {
            throw new ConfigurationException("No subdomain found for the added container : " + removedContainer);
        }

        for (Container jaxbContainer : jaxbSubdomainOfTheRemovedContainer.getContainer()) {
            if (jaxbContainer.getName().equals(removedContainer.getName())) {
                jaxbSubdomainOfTheRemovedContainer.getContainer().remove(jaxbContainer);
                break;
            }
        }
        
        this.containersConfig.remove(removedContainer.getName());
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void addSubdomainConfiguration(SubDomainConfiguration addedSubdomain) throws ConfigurationException {
        if (addedSubdomain == null) {
            throw new ConfigurationException("Can't add a null subdomain to the local topology");
        }

        if (this.subDomainsConfig.containsKey(addedSubdomain.getName())) {
            throw new ConfigurationException("Subdomain already registered for name : " + addedSubdomain.getName());
        }
        else {
            this.subDomainsConfig.put(addedSubdomain.getName(), addedSubdomain);

            //Add the new subdomain to the JAXB topology holder
            Subdomain jaxbSubdomain = new Subdomain();

            jaxbSubdomain.setName(addedSubdomain.getName());
            jaxbSubdomain.setDescription(addedSubdomain.getDescription());
            jaxbSubdomain.setMode(addedSubdomain.getMode());
            jaxbSubdomain.setNetworkTimeSynchronized(addedSubdomain.isNetworkTimeSynchronized());
            
            this.topology.getDomain().getSubDomain().add(jaxbSubdomain);
        }
    }

    public synchronized void removeSubdomainConfiguration(String removedSubdomainName) throws ConfigurationException {
        if (removedSubdomainName == null) {
            throw new ConfigurationException("Can't remove subdomain : null");
        }
        
        if (! this.subDomainsConfig.containsKey(removedSubdomainName)) {
            throw new ConfigurationException("No subdomain found for name : " + removedSubdomainName);
        }

        Subdomain removedJaxbSubdomain = null;
        for (Subdomain jaxbSubdomain : this.topology.getDomain().getSubDomain()) {
            if (jaxbSubdomain.getName().equals(removedSubdomainName)) {
                this.topology.getDomain().getSubDomain().remove(jaxbSubdomain);
                removedJaxbSubdomain = jaxbSubdomain;
                break;
            }
        }

        if (removedJaxbSubdomain == null) {
            throw new ConfigurationException("No subdomain found for name : " + removedSubdomainName);
        }

        this.subDomainsConfig.remove(removedSubdomainName);
        this.removeSubdomainContainers(removedSubdomainName);
    }

    private synchronized void removeSubdomainContainers(String removedSubdomainName) throws ConfigurationException {

        for (Map.Entry<String, ContainerConfiguration> containerEntry : this.containersConfig.entrySet()) {
            if (containerEntry.getValue().getSubdomainName().equals(removedSubdomainName)) {
                this.containersConfig.remove(containerEntry.getKey());
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized SubDomainConfiguration getSubDomainConfiguration() {
        this.log.call();
        return this.subDomainConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized DomainConfiguration getDomainConfiguration() {
        this.log.call();
        return this.domainConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Topology getTopology() throws ConfigurationException {
        this.log.call();
        // NOTE : This is the initial topology, not updated at runtime.
        return this.topology;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void setContainerState(ContainerState state) {
        this.containerConfiguration.setState(state);
    }

    /*
     * (non-Javadoc)
     *
     * @seeorg.ow2.petals.kernel.configuration.ConfigurationService#
     * verifyContainerTopology
     * (org.ow2.petals.kernel.configuration.ContainerConfiguration,
     * org.ow2.petals.kernel.configuration.SubDomainConfiguration,
     * org.ow2.petals.kernel.configuration.DomainConfiguration)
     */
    public synchronized void verifyContainerTopology(final ContainerConfiguration newContainerConfiguration,
            final SubDomainConfiguration newSubDomainConfiguration,
            final DomainConfiguration newDomainConfiguration) throws TopologyException {
        this.log.start();

        this.verifyDomain(newDomainConfiguration);

        this.verifySubDomain(newSubDomainConfiguration);

        this.verifyContainer(newContainerConfiguration);

        this.log.end();
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        try {
            // retrieve resources
            URL serverPropsURL = this.getClass().getResource(SERVER_PROPS_FILE);
            if (serverPropsURL == null) {
                throw new IOException("Failed to reach the resource '" + SERVER_PROPS_FILE + "'");
            }

            File serverPropsFile = new File(serverPropsURL.toURI().normalize());

            this.containersConfig = new HashMap<String, ContainerConfiguration>();
            this.subDomainsConfig = new HashMap<String, SubDomainConfiguration>();

            this.containerConfiguration = new ContainerConfiguration();
            this.containerConfiguration.init();
            this.subDomainConfiguration = new SubDomainConfiguration();
            this.domainConfiguration = new DomainConfiguration();

            this.createConfiguration(serverPropsFile);

        } catch (IOException e) {
            this.log.error("Failed to start configuration service", e);
            throw e;
        } catch (ConfigurationException e) {
            this.log.error("Failed to start configuration service", e);
            throw e;
        } catch (JAXBException e) {
            this.log.error("Failed to start configuration service", e);
            throw e;
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public synchronized ContainerConfiguration getContainerConfiguration(String containerName) {
        ContainerConfiguration result = null;
        if (this.containerConfiguration.getName().equals(containerName)) {
            result = this.getContainerConfiguration();
        } else {
            result = this.containersConfig.get(containerName);
        }
        return result;
    }

    /**
     * Create the container configuration based on the topology and server
     * files. Basic consistency is controlled between the 2 configuration files.
     * The {@code serverPropsFile} and {@code topologyFile} fields must be non
     * null.
     *
     * @param topologyFile
     *            The topology.xml file
     * @param serverPropsFile
     *            The server.properties file
     *
     * @throws JAXBException
     * @throws TopologyException
     * @throws IOException
     * @throws ConfigurationException
     *             Configuration consistency error
     */
    private synchronized void createConfiguration(File serverPropsFile) throws JAXBException, IOException,
            ConfigurationException {
        this.log.start();

        // retrieve the server properties
        this.serverProperties = new Properties();

        FileInputStream fis = new FileInputStream(serverPropsFile);
        try {
            this.serverProperties.load(fis);

            String containerName = this.serverProperties.getProperty(PROPERTY_CONTAINER_NAME);
            if (containerName == null) {
                containerName = System.getProperty(PROPERTY_CONTAINER_NAME);
            }

            this.loadTopology();

            final SubDomainAndContainerNames values = this.validateTopology(containerName);

            this.fillGlobalTopology();

            this.fillSubDomainTopology(values.subDomainName);

            this.fillLocalTopology(values.containerName);

            this.fillLocalProperties();

            this.fillGlobalContainerConfig();

            this.fillGlobalDomainConfig();

            // in standalone mode the container is always started
            if (org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE.equals(this.domainConfiguration.getMode())) {
                this.containerConfiguration.setState(ContainerState.STARTED);
            } else {
                // post validate platform configuration
                this.postValidate();
            }

            if (this.domainConfiguration.getMode() == org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE) {
                throw new ConfigurationException("Standalone mode is not supported anymore");
            }

            if (this.log.isDebugEnabled()) {
                this.log.debug("Configuration used :");
                this.log.debug(this.containerConfiguration);
            }
        } finally {
            fis.close();
        }

        this.log.end();
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void loadTopology() throws ConfigurationException {
        if (this.serverProperties == null) {
            throw new ConfigurationException("server properties are null");
        }

        File topologyFile = null;

        String topologyURL = this.serverProperties.getProperty(PROPERTY_TOPOLOGY_URL);
        // look if the topology field is defined at the server level
        if ((topologyURL != null) && (topologyURL.trim().length() > 0)) {
            // if local copy it to work, directory, if remote try to download
            // it...
            try {
                URL url = new URL(topologyURL);
                // get the file remotely
                File workFolder = new File(this.containerConfiguration.getWorkDirectoryPath());
                topologyFile = new File(workFolder, TOPOLOGY_FILE);
                FileUtils.copyURLToFile(url, topologyFile);

            } catch (MalformedURLException e) {
                throw new ConfigurationException("Bad topology URL '" + topologyURL + "'");
            } catch (IOException e) {
                throw new ConfigurationException("Can copy topology file from '" + topologyURL + "'");
            }

        } else {
            // get the file from the classpath
            URL classPathURL = this.getClass().getResource(TOPOLOGY_RESOURCE);
            if (classPathURL == null) {
                throw new ConfigurationException("Failed to get resource '" + TOPOLOGY_RESOURCE + "'");
            }

            try {
                topologyFile = new File(classPathURL.toURI().normalize());
            } catch (URISyntaxException e) {
                throw new ConfigurationException(e.getMessage());
            }
        }

        // topology file should be non null...
        try {
            this.topology = TopologyBuilder.createTopology(topologyFile);
        } catch (TopologyException e) {
            if (e.getCause() != null) {
                throw new ConfigurationException(e);
            } else {
                throw new ConfigurationException(e.getMessage());
            }
        }
    }

    private void fillGlobalContainerConfig() {
        Collection<Container> containers = TopologyHelper.findAllContainers(this.topology);
        for (Container container : containers) {
            ContainerConfiguration cc = new ContainerConfiguration();
            ContainerConfigurationHelper.fill(cc, container);
            cc.setSubdomainName(TopologyHelper.findSubdomain(container, this.topology).getName());
            this.containersConfig.put(cc.getName(), cc);
        }
    }

    private void fillGlobalDomainConfig() {
        List<Subdomain> list = this.topology.getDomain().getSubDomain();
        for (Subdomain subdomain : list) {
            SubDomainConfiguration subDomainCfg = new SubDomainConfiguration();
            subDomainCfg.setDescription(subdomain.getDescription());
            subDomainCfg.setName(subdomain.getName());
            subDomainCfg.setMode(subdomain.getMode());
            this.subDomainsConfig.put(subDomainCfg.getName(), subDomainCfg);
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Set<ContainerConfiguration> getContainersConfiguration() {
        Set<ContainerConfiguration> result = new HashSet<ContainerConfiguration>(
                this.containersConfig.size());
        for (ContainerConfiguration containerCfg : this.containersConfig.values()) {
            result.add(containerCfg);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Set<SubDomainConfiguration> getSubDomainsConfiguration() {
        Set<SubDomainConfiguration> result = new HashSet<SubDomainConfiguration>(
                this.subDomainsConfig.size());
        for (SubDomainConfiguration subDomainCfg : this.subDomainsConfig.values()) {
            result.add(subDomainCfg);
        }
        return result;
    }

    /**
     * Validate the consistency of the topology. The {@code topology} and the
     * {@code containerConfiguration} fields must be non {@code null}. Moreover,
     * if the domain is a standalone domain, {@code containerName} and {@code
     * subDomainName} are updated according to the topology.
     *
     * @param containerName
     *            The name of the local container retrieved from the
     *            environment, can be {@code null}
     * @return The container name and the sub-domain name of the provided
     *         container.
     * @throws ConfigurationException
     */
    private SubDomainAndContainerNames validateTopology(final String containerName)
            throws ConfigurationException {
        this.log.start();

        final SubDomainAndContainerNames resultValues = new SubDomainAndContainerNames();

        final Domain domain = this.topology.getDomain();
        final List<Subdomain> subdomainList = domain.getSubDomain();

        //Validate standalone topology
        if (domain.getMode().equals(DomainMode.STANDALONE)) {

            //Checks the number of subdomains
            if (subdomainList.size() > 1) {
                throw new ConfigurationException(
                        "Standalone topology mode accepts only one subdomain definition");
            }
            //At this point we have only one subdomain

            final Subdomain standaloneSubdomain = subdomainList.get(0);

            //Checks the number of containers for the only subdomain
            if (standaloneSubdomain.getContainer().size() > 1) {
                throw new ConfigurationException(
                        "Standalone topology mode accepts only one container definition");
            }
            //At this point we have only one container in only one subdomain

            //Check if the subdomain mode is standalone
            if (!standaloneSubdomain.getMode().equals(SubdomainMode.STANDALONE)) {
                throw new ConfigurationException(
                        "Standalone topology mode accepts only standalone subdomains");
            }
            //At this point we have a standalone subdomain and one container

            final Container standaloneContainer = standaloneSubdomain.getContainer().get(0);

            if (!standaloneContainer.getType().equals(NodeType.STANDALONE)) {
                throw new ConfigurationException(
                        "Standalone topology moded accepts only one container");
            }

            resultValues.containerName = standaloneContainer.getName();
        } //Validate the STATIC topology
        else if (domain.getMode().equals(DomainMode.STATIC) || domain.getMode().equals(DomainMode.DYNAMIC)) {

            //Checks if the container name has been set in the server.properties file
            if (containerName == null) {
                throw new ConfigurationException(
                        "The property " + PROPERTY_CONTAINER_NAME + " must be set in the server.properties file while being in distributed mode");
            }
            //We know the container name

            //Checks if the current container name can be found in the topology.xml file
            if (TopologyHelper.findContainer(containerName, this.topology) == null) {
                throw new ConfigurationException(
                        "Cannot find the definition for the container " + containerName + " in the topology.xml file");
            }
            //the container name has been found in the topology.xml file

            this.validateStaticTopology();

            resultValues.containerName = containerName;

        } //Woops :-)
        else {
            throw new ConfigurationException(
                    "Unsupported domain mode");
        }

        resultValues.subDomainName = TopologyHelper.findSubdomain(
                TopologyHelper.findContainer(resultValues.containerName, this.topology),
                this.topology).getName();

        // validate JNDI consistency
        if (this.topology.getDomain().getJndi() != null) {
            if ("".equals(this.topology.getDomain().getJndi().getFactory().trim())) {
                throw new ConfigurationException(
                        "JNDI factory classname must be set to reference an external JNDI server");
            }
            if (this.topology.getDomain().getJndi().getProviderUrl() == null) {
                throw new ConfigurationException(
                        "JNDI Provider URL must be set as an JNDI server factory is set '");
            }
        }



        // validate containers configuration against the domain mode
        if (!DomainMode.STANDALONE.equals(this.topology.getDomain().getMode())) {
            Collection<Container> containers = TopologyHelper.findAllContainers(this.topology);
            for (Container container : containers) {
                if (container.getTransportService() == null) {
                    throw new ConfigurationException(
                            "Transport service must be defined in the container '" + container.getName() + "' configuration");
                }
            }
        }

        //Yeap I know code below is dirty and inefficient...the whole topology / configuration needs to be heavily refactored
        for (Subdomain subdomain : subdomainList) {
            SubdomainMode subdomainMode = subdomain.getMode();

            boolean hasMasterNode = false;

            List<Container> subdomainContainersList = subdomain.getContainer();
            for (Container container : subdomainContainersList) {

                NodeType containerType = container.getType();
                if (subdomainMode.equals(SubdomainMode.FLOODING) && !containerType.equals(NodeType.PEER)) {
                    StringBuilder errorSB = new StringBuilder();
                    errorSB.append("When a subdomain is in flooding mode, all its ");
                    errorSB.append("containers must be of \"peer\" type");
                    throw new ConfigurationException(errorSB.toString());
                }

                if (subdomainMode.equals(SubdomainMode.STANDALONE) && !containerType.equals(NodeType.STANDALONE)) {
                    StringBuilder errorSB = new StringBuilder();
                    errorSB.append("When a subdomain is in standalone mode, the container ");
                    errorSB.append("must be in of \"standalone\" type");
                    throw new ConfigurationException(errorSB.toString());
                }

                if (subdomainMode.equals(SubdomainMode.MASTER_SLAVE) &&
                        !(containerType.equals(NodeType.MASTER) || containerType.equals(NodeType.SLAVE))) {
                    StringBuilder errorSB = new StringBuilder();
                    errorSB.append("When a subdomain is in master-slave mode, ");
                    errorSB.append("all its container must be either of ");
                    errorSB.append("\"master\" or \"slave\" type");
                    throw new ConfigurationException(errorSB.toString());
                }

                if (containerType.equals(NodeType.MASTER)) {
                    if (hasMasterNode) {
                        StringBuilder errorSB = new StringBuilder();
                        errorSB.append("Master node already defined for this subdomain");
                        throw new ConfigurationException(errorSB.toString());
                    } else {
                        hasMasterNode = true;
                    }
                }
            }

            if ((!hasMasterNode) && subdomainMode.equals(SubdomainMode.MASTER_SLAVE)) {
                StringBuilder errorSB = new StringBuilder();
                errorSB.append("No master node found for the current master-slave subdomain");
                throw new ConfigurationException(errorSB.toString());
            }
        }

        this.log.end();
        return resultValues;
    }

    /**
     * Validate the consistency of the static topology
     *
     * @throws ConfigurationException
     */
    private void validateStaticTopology() throws ConfigurationException {
        this.log.start();

        // Check all the sub domains names
        Collection<String> subdomainNameList = new HashSet<String>();
        for (Subdomain subdomain : this.topology.getDomain().getSubDomain()) {
            if (!subdomainNameList.add(subdomain.getName())) {
                throw new ConfigurationException("Duplicate sub-domain name '" + subdomain.getName() + "'");
            }
        }

        // Check all the containers
        Collection<String> nameList = new HashSet<String>();
        Map<String, HashSet<Integer>> portMap = new Hashtable<String, HashSet<Integer>>();
        for (Container container : TopologyHelper.findAllContainers(this.topology)) {
            if (container.getHost() == null) {
                throw new ConfigurationException("The container named '" + container.getName() + "' must declare its host element in static distributed mode");
            }

            HashSet<Integer> portList;
            if (!portMap.containsKey(container.getHost())) {
                portList = new HashSet<Integer>();
                portMap.put(container.getHost(), portList);
            } else {
                portList = portMap.get(container.getHost());
            }

            if ((container.getWebserviceService() != null) && (container.getWebserviceService().getPort() != 0)) {
                if (!portList.add(container.getWebserviceService().getPort())) {
                    throw new ConfigurationException("Duplicate port " + container.getWebserviceService().getPort() + " on host '" + container.getHost() + "'");
                }
            }

            if ((container.getRegistryService() != null) && (container.getRegistryService().getPort() != null)) {
                if (!portList.add(container.getRegistryService().getPort())) {
                    throw new ConfigurationException("Duplicate port " + container.getWebserviceService().getPort() + " on host '" + container.getHost() + "'");
                }
            }

            if (container.getJmxService().getRmiPort() != 0) {
                if (!portList.add(container.getJmxService().getRmiPort())) {
                    throw new ConfigurationException("Duplicate port " + container.getJmxService().getRmiPort() + " on host '" + container.getHost() + "'");
                }
            }

            if ((container.getTransportService() != null) && (container.getTransportService().getTcpPort() >= 0)) {
                if (! portList.add(container.getTransportService().getTcpPort())) {
                    throw new ConfigurationException("Duplicate port " + container.getTransportService().getTcpPort() + " on host '" + container.getHost() + "'");
                }
            }

            if (!nameList.add(container.getName())) {
                throw new ConfigurationException("Duplicate container name '" + container.getName() + "'");
            }
        }

        this.log.end();
    }

    /**
     * Fill the global topology.
     *
     * @throws ConfigurationException
     */
    private void fillGlobalTopology() throws ConfigurationException {
        this.log.start();

        // domain mode
        this.domainConfiguration.setName(this.topology.getDomain().getName());
        this.domainConfiguration.setDescription(this.topology.getDomain().getDescription());
        if (DomainMode.STANDALONE.equals(this.topology.getDomain().getMode())) {
            this.domainConfiguration.setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE);
            // set the default value of the routing strategy of standalone mode
            this.containerConfiguration.setRouterStrategy(RouterService.DEFAULT_STANDALONE_STRATEGY);
        } else if (DomainMode.STATIC.equals(this.topology.getDomain().getMode())) {
            this.domainConfiguration.setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STATIC);
        } else if (DomainMode.DYNAMIC.equals(this.topology.getDomain().getMode())) {
            this.domainConfiguration.setMode(org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.DYNAMIC);
        } else {
            throw new ConfigurationException("Unknown domain mode");
        }

        // domain JNDI
        if (this.topology.getDomain().getJndi() != null) {
            JndiConfiguration jndiConfiguration = new JndiConfiguration();

            jndiConfiguration.setJndiFactory(this.topology.getDomain().getJndi().getFactory());
            try {
                jndiConfiguration.setJndiProviderUrl(new URI(this.topology.getDomain().getJndi().getProviderUrl()));
            } catch (URISyntaxException e) {
                throw new ConfigurationException(e);
            }
            jndiConfiguration.setJndiSecurityPrincipal(this.topology.getDomain().getJndi().getSecurityPrincipal());
            jndiConfiguration.setJndiSecurityCredentials(this.topology.getDomain().getJndi().getSecurityCredentials());

            jndiConfiguration.setJndiPoolSize(this.topology.getDomain().getJndi().getPoolSize());
            jndiConfiguration.setJndiBatchSize(this.topology.getDomain().getJndi().getBatchSize());

            this.domainConfiguration.setJndiConfiguration(jndiConfiguration);
        }

        this.log.end();
    }

    /**
     * Fill the sub-domain configuration with its topology part. The {@code
     * subDomainConfiguration} and {@code topology} fields must be non {@code
     * null}.
     *
     * @param subDomainName
     *            The sub-domain name. Must be non null.
     * @throws ConfigurationException
     *             Configuration consistency error
     */
    private void fillSubDomainTopology(final String subDomainName) throws ConfigurationException {
        this.log.start();

        Subdomain subDomain = TopologyHelper.findSubdomain(subDomainName, this.topology);

        if (subDomain == null) {
            throw new ConfigurationException(
                    "Failed to retrieve the sub-domain configuration for the sub-domain named '" + subDomainName + "'");
        }

        this.subDomainConfiguration.setName(subDomainName);
        this.subDomainConfiguration.setDescription(subDomain.getDescription());
        if (subDomain.isNetworkTimeSynchronized() != null) {
            this.subDomainConfiguration.setNetworkTimeSynchronized(subDomain.isNetworkTimeSynchronized().booleanValue());
        }

        this.log.end();
    }

    /**
     * Fill the container configuration with its topology part. The {@code
     * containerConfiguration} and {@code topology} fields must be non {@code
     * null}.
     *
     * @param container
     *            name The local container name
     * @throws ConfigurationException
     *             Configuration consistency error
     */
    private void fillLocalTopology(String containerName) throws ConfigurationException {
        this.log.start();

        Container localContainer = TopologyHelper.findContainer(containerName, this.topology);

        if (localContainer == null) {
            throw new ConfigurationException(
                    "Failed to retrieve the container configuration for the container named '" + containerName + "'");
        }

        ContainerConfigurationHelper.fill(this.containerConfiguration, localContainer);
        this.containerConfiguration.setName(containerName);
        this.containerConfiguration.setSubdomainName(TopologyHelper.findSubdomain(localContainer,
                this.topology).getName());

        this.log.end();
    }

    /**
     * Fill the container configuration properties with its global properties
     * part. The {@code containerConfiguration} and {@code serverProperties}
     * fields must be non {@code null}.
     */
    private void fillLocalProperties() {
        this.log.start();
        Map<String, String> userProperties = new HashMap<String, String>();
        for (Object key : this.serverProperties.keySet()) {
            if (key instanceof String) {
                if (((String) key).equals(PROPERTY_REPOSITORY_PATH)) {
                    this.containerConfiguration.setRepositoryDirectoryPath(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_EXCHANGE_VALIDATION)) {
                    this.containerConfiguration.setExchangeValidation(Boolean.parseBoolean(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_QOS)) {
                    this.containerConfiguration.setTransportQOS(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_RECEIVERS)) {
                    this.containerConfiguration.setTCPReceivers(Integer.parseInt(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_SENDERS)) {
                    this.containerConfiguration.setTCPSenders(Integer.parseInt(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_CONNECTION_TIMEOUT)) {
                    this.containerConfiguration.setTCPConnectionTimeout(Long.parseLong(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_SEND_TIMEOUT)) {
                    this.containerConfiguration.setTCPSendTimeout(Long.parseLong(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_SEND_EVICTOR_DELAY)) {
                    this.containerConfiguration.setTCPSenderEvictorDelay(Long.parseLong(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TRANSPORT_TCP_SEND_EVICTABLE_DELAY)) {
                    this.containerConfiguration.setTCPSenderEvictableDelay(Long.parseLong(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_ROUTER_STRATEGY)) {
                    this.containerConfiguration.setRouterStrategy(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_ROUTER_SEND_ATTEMPT)) {
                    this.containerConfiguration.setRouterSendAttempt(Short.parseShort(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_ROUTER_SEND_DELAY)) {
                    this.containerConfiguration.setRouterSendDelay(Integer.parseInt(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_TASK_TIMEOUT)) {
                    this.containerConfiguration.setTaskTimeout(Long.parseLong(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_CLASSLOADER_ISOLATED)) {
                    this.containerConfiguration.setIsolateJBIClassLoaders(Boolean.parseBoolean(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_AUTOLOADER)) {
                    this.containerConfiguration.setActivateAutoloader(Boolean.parseBoolean(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_SSL_KEY_PASSWORD)) {
                    this.containerConfiguration.setSSLKeyPassword(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_SSL_KEYSTORE_FILE)) {
                    this.containerConfiguration.setSSLKeystore(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_SSL_KEYSTORE_PASSWORD)) {
                    this.containerConfiguration.setSSLKeystorePassword(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_SSL_TRUSTSTORE_FILE)) {
                    this.containerConfiguration.setSSLTruststore(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_SSL_TRUSTSTORE_PASSWORD)) {
                    this.containerConfiguration.setSSLTruststorePassword(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_TRANSPORT_TIMEOUT)) {
                    this.containerConfiguration.setRegistryTransportTimeout(Integer.parseInt(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_REGISTRY_SYNCHRO_PERIOD)) {
                    this.containerConfiguration.setRegistrySynchroPeriod(Integer.parseInt(this.serverProperties.getProperty((String) key)));
                } else if (((String) key).equals(PROPERTY_REGISTRY_INCOMING_MANAGER)) {
                    this.containerConfiguration.setRegistryIncomingManagerClassName(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_OUTGOING_MANAGER)) {
                    this.containerConfiguration.setRegistryOutgoingManagerClassName(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_REPOSITORY)) {
                    this.containerConfiguration.setRegistryRepository(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_MESSAGE_RECEIVER)) {
                    this.containerConfiguration.setRegistryMessageReceiverClassName(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_MESSAGE_SENDER)) {
                    this.containerConfiguration.setRegistryMessageSenderClassName(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_DB_DRIVER)) {
                    this.containerConfiguration.setRegistryDatabaseDriver(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_DB_URL)) {
                    this.containerConfiguration.setRegistryDatabaseUrl(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_DB_USER)) {
                    this.containerConfiguration.setRegistryDatabaseUser(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_DB_PASSWORD)) {
                    this.containerConfiguration.setRegistryDatabasePassword(this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(PROPERTY_REGISTRY_DB_DIALECT)) {
                    this.containerConfiguration.setRegistryDatabaseDialect(this.serverProperties.getProperty((String) key));
                } else if (((String) key).startsWith(PROPERTY_USER_CONFIG_PREFIX + ".")) {
                    userProperties.put((String) key, this.serverProperties.getProperty((String) key));
                } else if (((String) key).equals(TOPOLOGY_UPDATE_PERIOD)) {
                    this.containerConfiguration.setTopologyUpdatePeriod(Long.parseLong(this.serverProperties.getProperty((String) key, "30")));
                }
            }
        }

        if (userProperties.size() > 0) {
            this.containerConfiguration.setUserConfiguration(userProperties);
        }

        this.log.end();
    }

    /**
     * Post validation on the configuration, once the topology and properties
     * are filled.
     *
     * @throws ConfigurationException
     */
    private void postValidate() throws ConfigurationException {
        if (RouterService.FAST_POLICY.equals(this.containerConfiguration.getRouterQOS()) && (this.containerConfiguration.getTCPPort() == 0)) {
            throw new ConfigurationException("The router transport QOS '" + RouterService.FAST_POLICY + "' is not compatible with the transport protocols configuration");
        }
        if (RouterService.SECURE_POLICY.equals(this.containerConfiguration.getRouterQOS())) {
            throw new ConfigurationException("The default router QOS '" + RouterService.SECURE_POLICY + "' is not compatible with the transport protocols configuration");
        }
    }

    /**
     * Verify the given new domain configuration against the topology.
     *
     * @param newDomainConfiguration
     * @throws TopologyException
     */
    private void verifyDomain(DomainConfiguration newDomainConfiguration) throws TopologyException {
        this.log.debug("Check the domain configuration of the new Container...");

        if (!newDomainConfiguration.getMode().equals(this.domainConfiguration.getMode())) {
            throw new TopologyException("Domain mode mismatch");
        }

        // JNDI service
        if (!StringHelper.equal(newDomainConfiguration.getJndiConfiguration().getJndiFactory(), this.domainConfiguration.getJndiConfiguration().getJndiFactory())) {
            throw new TopologyException("JNDI factory mismatch");
        }
        if (((newDomainConfiguration.getJndiConfiguration().getJndiProviderUrl() == null) && (this.domainConfiguration.getJndiConfiguration().getJndiProviderUrl() != null)) || ((newDomainConfiguration.getJndiConfiguration().getJndiProviderUrl() != null) && (this.domainConfiguration.getJndiConfiguration().getJndiProviderUrl() == null)) || ((newDomainConfiguration.getJndiConfiguration().getJndiProviderUrl() != null) && (this.domainConfiguration.getJndiConfiguration().getJndiProviderUrl() != null) && newDomainConfiguration.getJndiConfiguration().getJndiProviderUrl().toString().equals(this.domainConfiguration.toString()))) {
            throw new TopologyException("JNDI Provider Url mismatch");
        }
        if (!StringHelper.equal(newDomainConfiguration.getJndiConfiguration().getJndiSecurityPrincipal(),
                this.domainConfiguration.getJndiConfiguration().getJndiSecurityPrincipal())) {
            throw new TopologyException("JNDI Security Principal mismatch");
        }
        if (!StringHelper.equal(newDomainConfiguration.getJndiConfiguration().getJndiSecurityCredentials(),
                this.domainConfiguration.getJndiConfiguration().getJndiSecurityCredentials())) {
            throw new TopologyException("JNDI Security Credentials mismatch");
        }
        if (newDomainConfiguration.getJndiConfiguration().getJndiPoolSize() != this.domainConfiguration.getJndiConfiguration().getJndiPoolSize()) {
            throw new TopologyException("JNDI Pool Size mismatch");
        }
        if (newDomainConfiguration.getJndiConfiguration().getJndiBatchSize() != this.domainConfiguration.getJndiConfiguration().getJndiBatchSize()) {
            throw new TopologyException("JNDI Batch Size mismatch");
        }

        this.log.debug("Domain configuration of the new Container is OK");
    }

    /**
     * Verify the given new sub-domain configuration against the topology.
     *
     * @param newDomainConfiguration
     * @throws TopologyException
     */
    private void verifySubDomain(final SubDomainConfiguration newSubDomainConfiguration)
            throws TopologyException {
        this.log.debug("Check the sub-domain configuration of the new Container...");

        if (DomainMode.STATIC.equals(this.topology.getDomain().getMode())) {
            Subdomain newSubdomain = TopologyHelper.findSubdomain(newSubDomainConfiguration.getName(), this.topology);

            if (newSubdomain == null) {
                throw new TopologyException("Subdomain of the Container unknown");
            }
            if (newSubdomain.isNetworkTimeSynchronized() == null) {
                if (newSubDomainConfiguration.isNetworkTimeSynchronized() == true) {
                    throw new TopologyException("Network synchronization mismatch");
                }
            } else if (newSubdomain.isNetworkTimeSynchronized().booleanValue() != newSubDomainConfiguration.isNetworkTimeSynchronized()) {
                throw new TopologyException("Network synchronization mismatch");
            }
        } else if (newSubDomainConfiguration.getName().equals(this.subDomainConfiguration.getName())) {
            if (newSubDomainConfiguration.isNetworkTimeSynchronized() != this.subDomainConfiguration.isNetworkTimeSynchronized()) {
                throw new TopologyException("Network synchronization mismatch");
            }
        }

        this.log.debug("Sub-domain configuration of the new Container is OK");
    }

    /**
     * Verify the given new container configuration against the topology.
     *
     * @param newContainerConfiguration
     * @throws TopologyException
     */
    private void verifyContainer(ContainerConfiguration newContainerConfiguration)
            throws TopologyException {

        this.log.debug("Check the container configuration of the new Container.");
        if (DomainMode.STATIC.equals(this.topology.getDomain().getMode())) {
            Container newContainer = TopologyHelper.findContainer(newContainerConfiguration.getName(), this.topology);

            if (newContainer == null) {
                throw new TopologyException("Container name unknown");
            }
            this.log.debug("Check container " + newContainer.getName() + " ...");
            if (!StringHelper.equal(newContainer.getUser(), newContainerConfiguration.getUser())) {
                throw new TopologyException("User mismatch");
            }
            if (!StringHelper.equal(newContainer.getPassword(), newContainerConfiguration.getPassword())) {
                throw new TopologyException("Password mismatch");
            }

            // JMX Service
            if (newContainer.getJmxService().getRmiPort() != newContainerConfiguration.getJmxRMIConnectorPort()) {
                throw new TopologyException("JMX RMI connector port mismatch");
            }

            //Transport service
            if (newContainer.getTransportService() == null) {
                throw new TopologyException("Can't find transport service configuration");
            }
            if (newContainer.getTransportService().getTcpPort() == 0) {
                throw new TopologyException("Tranport service TCP port not set");
            }
            if (newContainer.getTransportService().getTcpPort() != newContainerConfiguration.getTCPPort()) {
                throw new TopologyException("TCP port mismatch");
            }
        } else if (newContainerConfiguration.getHost().equals(this.containerConfiguration.getHost())) {
            if (newContainerConfiguration.getName().equals(this.containerConfiguration.getName())) {
                throw new TopologyException("Duplicate container name '" + this.containerConfiguration.getName() + "'");
            }

        }

        this.log.debug("Container configuration of the new Container is OK");
    }

    public void validateDynamicTopology() throws ConfigurationException {
        try {
            this.verifyDomain(this.domainConfiguration);

            for (Map.Entry<String, SubDomainConfiguration> subdomainEntry : this.subDomainsConfig.entrySet()) {
                this.verifySubDomain(subdomainEntry.getValue());
            }

            for (Map.Entry<String, ContainerConfiguration> containerEntry : this.containersConfig.entrySet()) {
                this.verifyContainer(containerEntry.getValue());
            }

        } catch (TopologyException ex) {
            throw new ConfigurationException(ex);
        }
    }

    public void backupTopology() throws ConfigurationException {

        if (this.serverProperties == null) {
            throw new ConfigurationException("server properties are null");
        }

        File topologyFile = null;

        String topologyURL = this.serverProperties.getProperty(PROPERTY_TOPOLOGY_URL);
        
        // look if the topology field is defined at the server level
        if ((topologyURL != null) && (topologyURL.trim().length() > 0)) {
            // if local copy it to work, directory, if remote try to download
            // it...
            try {
                URL url = new URL(topologyURL);
                // get the file remotely
                File workFolder = new File(this.containerConfiguration.getWorkDirectoryPath());
                topologyFile = new File(workFolder, TOPOLOGY_FILE);
                FileUtils.copyURLToFile(url, topologyFile);

            } catch (MalformedURLException e) {
                throw new ConfigurationException("Bad topology URL '" + topologyURL + "'");
            } catch (IOException e) {
                throw new ConfigurationException("Can copy topology file from '" + topologyURL + "'");
            }

        } else {
            // get the file from the classpath
            URL classPathURL = this.getClass().getResource(TOPOLOGY_RESOURCE);
            if (classPathURL == null) {
                throw new ConfigurationException("Failed to get resource '" + TOPOLOGY_RESOURCE + "'");
            }

            try {
                topologyFile = new File(classPathURL.toURI().normalize());
            } catch (URISyntaxException e) {
                e.printStackTrace();
                throw new ConfigurationException(e.getMessage());
            }
        }

        // topology file should be non null...
        try {
            TopologyBuilder.marshallTopologyFile(topologyFile, this.topology);
        } catch (TopologyException e) {
            throw new ConfigurationException(e.getMessage());
        }
    }
}
