/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.transport.local;

import java.util.ConcurrentModificationException;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.jbi.messaging.MessagingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.exchange.MessageExchange;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.TransportListener;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.util.TransportSendContext;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.util.LoggingUtil;

/**
 * @author Roland NAUDIN - EBM WebSourcing
 * 
 */
@FractalComponent
@Provides(interfaces = { @Interface(name = "service", signature = org.ow2.petals.transport.Transporter.class) })
public class LocalTransporterImpl implements Transporter {

    /**
     * The maximum size of a the Java queue.
     */
    private static final int QUEUE_SIZE = 10000;

    /**
     * Monolog component logger
     */
    @Monolog(name = "logger")
    private Logger logger;

    /**
     * the Transporter listener (router) Fractal component
     */
    @Requires(name = "transportlistener", signature = TransportListener.class)
    private TransportListener transportListener;

    /**
     * The logger.
     */
    private LoggingUtil log;

    /**
     * List of exchanges that are blocked during a synchronous send
     */
    private List<MessageExchange> pendingSyncExchanges;

    /**
     * Traffic flag to handle stop of traffic
     */
    private boolean stopTraffic;

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#prepareStop()
     */
    public void stopTraffic() {
        this.log.start();
        this.stopTraffic = true;

        // Awake all the pending 'SYNC_SEND' issuer threads to abort the
        // invocation
        boolean redo = true;
        while (redo) {
            try {
                for (MessageExchange messageExchangeImpl : this.pendingSyncExchanges) {
                    synchronized (messageExchangeImpl) {
                        messageExchangeImpl.notify();
                    }
                }
                redo = false;
            } catch (ConcurrentModificationException e) {
                // Try again
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#send(org.ow2.petals.jbi.messaging
     * .exchange.MessageExchangeImpl,
     * org.ow2.petals.transport.util.TransportSendContext)
     */
    public void send(final MessageExchange exchange, final TransportSendContext transportContext)
            throws TransportException {
        this.log.start();

        this.checkTransporterState();

        if (TransporterUtil.getSyncMode(exchange, true)) {
            this.pendingSyncExchanges.remove(exchange);
            synchronized (exchange) {
                exchange.notify();
            }
        } else {
            this.transportListener.onExchange(exchange);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#sendSync(org.ow2.petals.jbi.messaging
     * .exchange.MessageExchangeImpl,
     * org.ow2.petals.transport.util.TransportSendContext)
     */
    public MessageExchange sendSync(final MessageExchange exchange,
            final TransportSendContext transportContext) throws TransportException {
        this.log.start();

        this.checkTransporterState();

        // set the relevant properties to mark the message as a
        // synchronous send
        boolean isOverSync;
        try {
            isOverSync = TransporterUtil.setSendSyncProperties(exchange);
        } catch (MessagingException e) {
            throw new TransportException(e);
        }

        synchronized (exchange) {
            if (isOverSync) {
                // in this case the current send sync is a response of a
                // previous send sync
                exchange.notify();
            } else {
                this.pendingSyncExchanges.add(exchange);
            }

            try {
                if (!isOverSync) {
                    this.transportListener.onExchange(exchange);
                }
                // wait for a response or a timeout reached
                exchange.wait(transportContext.timeout);
            } catch (InterruptedException e) {
                throw new TransportException(
                        "Failed to receive the response message of a synchronous send", e);
            }
        }

        // reset the pending exchange it is an 'overSync' exchange
        if (!TransporterUtil.updateSyncProperties(exchange)
                && this.pendingSyncExchanges.contains(exchange)) {
            // If exchange is in the pending list,
            // it means that the timeout has been reached or the transporter is
            // stopping
            this.checkTransporterState();

            // try to remove the exchange if a timeout
            this.pendingSyncExchanges.remove(exchange);

            this.log.warning("Failed to send synchronously the exchange: "
                    + exchange.getExchangeId() + ". Timeout occured");

            this.log.end();
            return null;
        }

        this.log.end();

        return exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.ProtocolMonitor#getQueueMaxSize()
     */
    public int getQueueMaxSize() throws TransportException {
        return QUEUE_SIZE;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.ProtocolMonitor#getQueueSizes()
     */
    public Map<String, Integer> getQueueSizes() throws TransportException {

        return null;
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.call();

        this.pendingSyncExchanges = new Vector<MessageExchange>(100);
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

    /**
     * Check if the transporter is running, if not, raise an exception.
     * 
     * @throws TransportException
     */
    private void checkTransporterState() throws TransportException {
        if (this.stopTraffic) {
            Thread.currentThread().interrupt();
            throw new TransportException("The Transporter traffic is stopped",
                    new InterruptedException());
        }
    }

}
