/**
 * PETALS - PETALS Services Platform. Copyright (c) 2009 OW2,
 * http://www.ow2.org/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.transport.platform.nio;

import java.io.IOException;
import java.util.ConcurrentModificationException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.jbi.messaging.MessagingException;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.exchange.MessageExchange;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.TransportListener;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.platform.nio.client.NioClientObjectPool;
import org.ow2.petals.transport.platform.nio.selector.NioClient;
import org.ow2.petals.transport.platform.nio.server.NioReceiver;
import org.ow2.petals.transport.util.TransportSendContext;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.util.LoggingUtil;

/**
 * NIO Implementation of the Transporter
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 */
@FractalComponent
@Provides(interfaces = { @Interface(name = "service", signature = Transporter.class),
        @Interface(name = NioReceiver.NIO_MSG_RECEIVER_ITF, signature = NioReceiver.class) })
public class NioTransporterImpl implements Transporter, NioReceiver {

    /**
     * Fractal Logger
     */
    @Monolog(name = "logger")
    private Logger logger;

    /**
     * the Transporter listener (router) Fractal component
     */
    @Requires(name = "transportlistener", signature = TransportListener.class)
    private TransportListener transportListener;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * All connections to remote containers are pooled into a keyed-pool. The
     * key is the name of the remote container.
     */
    @Requires(name = NioClientObjectPool.NIO_CLIENT_OBJECT_ITF, signature = NioClientObjectPool.class)
    private NioClientObjectPool clientConnections;

    /**
     * Traffic flag to handle stop of traffic
     */
    private boolean stopTraffic;

    /**
     * Map of exchanges that are blocked during a synchronous send
     */
    private Map<String, MessageExchange> pendingSyncExchanges;

    /**
     * Logger Wrapper
     */
    private LoggingUtil log;

    /**
     * Timeout to send a message exchange.
     */
    private long sendTimeout;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.platform.nio.server.NioReceiver#onMessage(java
     * .lang.Object)
     */
    public void onMessage(final Object object) {
        this.log.start();

        if (object instanceof MessageExchange) {
            final MessageExchange messageExchange = (MessageExchange) object;

            if (TransporterUtil.getSyncMode(messageExchange, true)) {
                if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                    this.log.debug("The message " + messageExchange.getExchangeId()
                            + " is a synchronized response");
                }
                TransporterUtil.updateSyncProperties(messageExchange);
                MessageExchange exchange = this.pendingSyncExchanges.put(messageExchange
                        .getExchangeId(), messageExchange);
                synchronized (exchange) {
                    exchange.notify();
                }
            } else {
                if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                    this.log.debug("The message " + messageExchange.getExchangeId()
                            + " is not a synchronized response");
                }
                this.transportListener.onExchange(messageExchange);
            }
        } else {
            // This case doesn't have to exist.
            this.log.warning("Unexpected received data. Data are discarded");
        }

        this.log.end();
    }

    /**
     * This method does following action :
     * <ol>
     * <li>Serialize the message exchange.</li>
     * <li>Get the destination of the message.</li>
     * <li>Add it to the message.</li>
     * <li>Send it to the transport agent.</li>
     * </ol>
     * 
     * @see Transporter#send(MessageExchangeImpl, java.lang.String,
     *      java.lang.String, long)
     */
    public void send(MessageExchange exchange, TransportSendContext transportContext)
            throws TransportException {
        this.log.start("Send exchange to destination '" + transportContext.destination + "'");

        this.checkTransporterState();

        try {
            final NioClient clientConnection = this.clientConnections
                    .borrowObject(transportContext.destination.getContainerName());

            try {
                clientConnection.send(exchange, this.sendTimeout);
                // In case of error during send, the clientConnection is
                // automatically invalidate.
            } finally {
                this.clientConnections.returnObject(
                        transportContext.destination.getContainerName(), clientConnection);
            }

        } catch (final IOException e) {
            throw new TransportException(e);
        } catch (final InterruptedException e) {
            throw new TransportException(e);
        } catch (final Exception e) {
            throw new TransportException(e);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#sendSync(org.ow2.petals.jbi.messaging
     * .exchange.MessageExchange,
     * org.ow2.petals.transport.util.TransportSendContext)
     */
    public MessageExchange sendSync(final MessageExchange exchange,
            final TransportSendContext transportContext) throws TransportException {
        this.log
                .start("Send synchronous exchange to destination : " + transportContext.destination);
        MessageExchange responseExchange;

        this.checkTransporterState();

        // set the relevant properties to mark the message as a
        // synchronous send
        try {
            TransporterUtil.setSendSyncProperties(exchange);
        } catch (final MessagingException e) {
            throw new TransportException(e);
        }

        this.pendingSyncExchanges.put(exchange.getExchangeId(), exchange);

        try {
            synchronized (exchange) {
                this.send(exchange, transportContext);
                // wait for a response or a timeout reached
                exchange.wait(transportContext.timeout);
            }
            responseExchange = this.pendingSyncExchanges.remove(exchange.getExchangeId());
        } catch (final InterruptedException e) {
            TransporterUtil.updateSyncProperties(exchange);
            this.pendingSyncExchanges.remove(exchange);
            throw new TransportException(
                    "Failed to receive the response message of a synchronous send", e);
        }

        // If responseExchange is the exchange,
        // it means that the timeout has been reached or the transporter is
        // stopping
        if (responseExchange == exchange) {
            this.checkTransporterState();

            this.log.warning("Failed to send synchronously the exchange: "
                    + exchange.getExchangeId() + ". Timeout occured");

            this.log.end();
            return null;
        }

        this.log.end();

        return responseExchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#stopTraffic()
     */
    public void stopTraffic() {
        this.log.start();
        this.stopTraffic = true;

        // Awake all the pending 'SYNC_SEND' issuer threads to abort the
        // invocation
        boolean redo = true;
        while (redo) {
            try {
                for (final MessageExchange messageExchangeImpl : this.pendingSyncExchanges.values()) {
                    synchronized (messageExchangeImpl) {
                        messageExchangeImpl.notify();
                    }
                }
                redo = false;
            } catch (final ConcurrentModificationException e) {
                // Try again
            }
        }

        this.log.end();
    }

    /**
     * Init variables with Fractal attributes value and create/bind the topic
     * for this container.
     * 
     * @throws NoSuchInterfaceException
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IllegalLifeCycleException, NoSuchInterfaceException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.sendTimeout = this.configurationService.getContainerConfiguration()
                .getTCPSendTimeout();
        this.pendingSyncExchanges = new ConcurrentHashMap<String, MessageExchange>(100);
    }

    /**
     * Method called when the fractal component is stopped
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws Exception {
        this.log.call();
    }

    /**
     * Check if the transporter is running, if not, raise an exception.
     * 
     * @throws TransportException
     */
    private void checkTransporterState() throws TransportException {
        if (this.stopTraffic) {
            Thread.currentThread().interrupt();
            throw new TransportException("The Transporter traffic is stopped",
                    new InterruptedException());
        }
    }

}
